#include <stdio.h>
#include <string.h>
#include <sys/time.h>
#include <assert.h>
#include <omp.h>
#include <stdlib.h>

#include "Common_openmp.h"

#define JUSTFRAGMENT 0

//Globals
static int FASTQINPUT = 1, FASTAINPUT = 0;
static char blank[6] = "BLANK", bases[BASES+1] = "ACGTN", complement[BASES+1] = "TGCAN", aa[AA+1] = "ABCDEFGHIKLMNPQRSTVWXYZ";
static float sub_matrix[BASES][BASES];
int subst_index[LETTERS];
float pos_specific_scores[MAXQLEN * BASES + BASES + PSS_OFFSET];

void compute_scorez_matrix(char* query, char* phred, float* phred_probs){ 

  int basecount = strlen(bases), i, j, k;
  if(FASTQINPUT) {
    for (i = 0; i < strlen(query); i++) {
      float score = 0;
      for (k = 0; k < basecount; k++) {
        score=0;
        for (j = 0; j < basecount; j++) {
          if(bases[j] == query[i]) score += phred_probs[i]*sub_matrix[subst_index[bases[k] - 'A']][j];
          else if(bases[j]=='A' || bases[j]=='C' || bases[j]=='G' || bases[j]=='T') score += (1/3)*(1-phred_probs[i])*sub_matrix[subst_index[bases[k] - 'A']][j];
        }
        pos_specific_scores[i*BASES + k + PSS_OFFSET] = score;
      }
    }
  }
  if(FASTAINPUT) {
    for (i = 0; i < strlen(query); i++) {
      float score = 0;
      for (k = 0; k < basecount; k++) {
        score=0;
        for (j = 0; j < basecount; j++) {
          if(bases[j] == query[i]) score += sub_matrix[subst_index[bases[k] - 'A']][j];
        }
        pos_specific_scores[i*BASES + k + PSS_OFFSET] = score;
      }
    }
  }

if(DEBUG > 1){
      for (i = 0; i < strlen(query); i++) {
        printf("%d: ", i);
        for (k = 0; k < basecount; k++) printf("%f ", pos_specific_scores[i*BASES + k + PSS_OFFSET]); printf("\n");
      }
      for (i = 0; i < basecount; i++) printf("%d ", subst_index[bases[i]-'A']); printf("\n");
      for (i = 0; i < LETTERS; i++) printf("%d ", subst_index[i]); printf("\n");
}

}

unsigned int get_dblength(int genomes, char dbfilenames[MAX_CONTIGS][FILENAME_LIMIT], int FRAGMENTSIZE) {

  struct timeval starttime, endtime; float seconds;

if(OUTPUTTIME) gettimeofday(&starttime, NULL);

  FILE* dbfile;
  unsigned int dblength=0;
  int i, j, l; 
  char *line = NULL; size_t len = 0; ssize_t read;

  for(i = 0; i < genomes; i++) {
    dbfile = fopen(dbfilenames[i], "r");
//    printf("Reading filename %s\n", dbfilenames[i]);
    if(dbfile == NULL) {
      printf("Error: No input file specified OR input file not found: file=%s\n", dbfilenames[i]);
      exit(EXIT_FAILURE);
    }
    while ((read = getline(&line, &len, dbfile)) != -1) {
      if(line[0] != '>')
        dblength += (read-1);
      else {
        if(dblength > 0 && dblength % FRAGMENTSIZE) {
          l = FRAGMENTSIZE - (dblength % FRAGMENTSIZE);
          for(j = 0; j < l; j++)
            dblength++;
        }
      }
    }
//    printf("dblength=%u\n", dblength);
  }
  if(dblength % FRAGMENTSIZE) {
    l = FRAGMENTSIZE - (dblength % FRAGMENTSIZE);
    for(j = 0; j < l; j++)
      dblength++;
  }
  if(line)
    free(line);

if(OUTPUTTIME){
  gettimeofday(&endtime, NULL); seconds=((double)endtime.tv_sec+(double)endtime.tv_usec/1000000)-((double)starttime.tv_sec+(double)starttime.tv_usec/1000000);}
if(OUTPUTTIME)
  printf("Time to get db size is %f\n", seconds);

//  printf("dblength=%u\n", dblength);
  return dblength;
}

void get_database(char* database, char* contignames, unsigned int  *contigposition, int genomes, char dbfilenames[MAX_CONTIGS][FILENAME_LIMIT], \
		  int FRAGMENTSIZE, int NUMFRAGMENTS, int FNAME) {

  struct timeval starttime, endtime; float seconds;

if(OUTPUTTIME) gettimeofday(&starttime, NULL);

  FILE* dbfile;
  int i, j, l;
  char *line = NULL, contigname[FNAME]; size_t len = 0; ssize_t read;
  unsigned int ui=0, prevui=0, uj=0; int fragmentid=0;

  for(i = 0; i < genomes; i++) {
    dbfile = fopen(dbfilenames[i], "r");
//    printf("Reading filename %s\n", dbfilenames[i]); 
    if(dbfile == NULL) {
      printf("Error: No input file specified\n");
      exit(EXIT_FAILURE);
    }
    while ((read = getline(&line, &len, dbfile)) != -1) {
      if(strcmp(line, "") == 0) continue;
      if(line[0] != '>') {
	//-1 for the end of line character
        for(j = 0; j < strlen(line)-1; j++) {
          database[ui] = line[j];
          if(database[ui] != 'A' && database[ui] != 'C' && database[ui] != 'G' &&
            database[ui] != 'T' && database[ui] != 'N'){
            printf("Character %c found in genome. Changing it to N\n", database[ui]);
            database[ui] = 'N';
          }
          ui++;
        }
      }
      else {
        if(ui > 0) {
          //Pad N's into database
          if(ui % FRAGMENTSIZE){
            l = FRAGMENTSIZE - (ui % FRAGMENTSIZE);
            for(j = 0; j < l; j++) 
              database[ui++] = 'N'; 
          }
          l = (ui - prevui)/FRAGMENTSIZE;
          uj = 0;
          for(j = 0; j < l; j++) {
            strcpy(&contignames[fragmentid*FNAME], contigname);
            contigposition[fragmentid]=uj;
  	    fragmentid++; uj += FRAGMENTSIZE;
          }
        }
        prevui = ui;
        strncpy(contigname, line, FNAME); 
        contigname[FNAME-1]='\0';
	//remove end of line
        contigname[strlen(contigname)-1]='\0';
        strcpy(&contignames[fragmentid*FNAME], contigname); 
      }
    }
//    printf("ui=%u fragmentid=%d\n", ui, fragmentid);
  }
  //Pad N's into database
  if(ui % FRAGMENTSIZE){
    l = FRAGMENTSIZE - (ui % FRAGMENTSIZE);
    for(j = 0; j < l; j++)
      database[ui++] = 'N';
  }
  l = (ui - prevui)/FRAGMENTSIZE;
  uj = 0;
  for(j = 0; j < l; j++) {
    strcpy(&contignames[fragmentid*FNAME], contigname);
    contigposition[fragmentid]=uj;
    fragmentid++; uj += FRAGMENTSIZE;
  }
  if(line)
    free(line);

//  printf("fragmentid=%d NUMFRAGMENTS=%d genomes=%d\n", fragmentid, NUMFRAGMENTS, genomes);
  assert(fragmentid == NUMFRAGMENTS);

if(OUTPUTTIME){
  gettimeofday(&endtime, NULL); seconds=((double)endtime.tv_sec+(double)endtime.tv_usec/1000000)-((double)starttime.tv_sec+(double)starttime.tv_usec/1000000);}
if(OUTPUTTIME)
   printf("Time to read genomes is %f\n", seconds);       

/* Testing code */
/*  printf("fragmentid=%d\n", fragmentid);
  ui=0;
  for(i = 0; i<fragmentid; i++){
    printf("i=%d pos=%d %s\n", i, contigposition[i], &contignames[i*FNAME]);
//    for(j=0; j<FRAGMENTSIZE; j++) printf("%c", database[ui++]);
//    printf("\n");
  } 
exit(EXIT_SUCCESS); 
*/

}

int DNAGenome ( int argc, char* argv[] ) {

//Locals
  int i, j, l, firstcounter, secondcounter, lencounter, count, genomes;
  unsigned int genomelength = 0, dblen, ui;
  float p1, p2, p3; short reverse = 0, reverse2 = 0;
  FILE *qfile; 
  int bestfragmentnumber=0, bestfragmentnumber2=0, fragmentsize=0, startpos=-1, endpos=-1, startalign=-1, endalign=-1;
  float bestscore=NEG_INF, bestscore2=NEG_INF, matches=0, mismatches=0, gaps=0;
  char *firstalign, *secondalign, *cigar, *cigarstr, qfilename[FILENAME_LIMIT], dbfilenames[MAX_CONTIGS][FILENAME_LIMIT];
  float GAPOPEN=DEFAULT_GAPOPEN;
  int FINALFRAGMENTSIZE = 0, QNAME = DEFAULT_QNAME, FNAME = DEFAULT_FNAME, JOBS = 0, THREADS = 256, CORES = 1; 
  int FRAGMENTSIZE = -1, QLEN = -1, NUMFRAGMENTS = -1, NUMQUERIES = -1, numfragments = -1, QUERY_PADLENGTH = -1;
  int leftfrags = 0, rightfrags = 0, start = 0;
  short SAMOUTPUT=1, samout=0;

  struct timeval starttime, endtime, start2time, end2time; float seconds;

  char cmdline[2000]; cmdline[0]='\0';
  for(i=1; i<argc; i++){ strcat(cmdline, argv[i]); strcat(cmdline, " "); }
  
  char help_message[3000] = 
"MaxSSmap version 1.0 (May 2013) maps DNA to genomes sequences \n\
with a maximum scoring subsequence strategy. MaxSSmap is written\n\
by Usman Roshan and Yumeng Wen.\n\n\
Usage: ./MaxSSmap dnagenome <options>\n\n\
 -q <file>       fastq file containing multiple reads\n\
 -f <file>       genome files in fasta format (unzipped)\n\
 -fl <n>         fragment length (must specify; must be a multiple of 16;\n\
	         recommend 48 for Ecoli and similar length genomes\n\
 -nq <n>         number of reads in fastq or fasta file (must specify)\n\
 -t <n>          number of threads (default 256)\n\
 -cores <n>      number of CPU cores (default 1)\n\
 -h              this help message\n\n\
Please contact usman@cs.njit.edu for support and additional information\n";

  qfilename[0]='\0'; dbfilenames[0][0]='\0'; 
//Get parameters
  if(argc == 2) { printf(help_message); exit(EXIT_SUCCESS); }
  for(i=0; i<argc; i++){
    if(!strcmp(argv[i], "-fl")) FRAGMENTSIZE = atoi(argv[i+1]);
    else if(!strcmp(argv[i], "-q")) strcpy(qfilename, argv[i+1]);
    else if(!strcmp(argv[i], "-f")) { l=0; i++; while(i<argc && argv[i][0] != '-') {strcpy(dbfilenames[l], argv[i]); l++; i++; } i--; genomes=l;}
    else if(!strcmp(argv[i], "-nq")) NUMQUERIES = atoi(argv[i+1]); 
    else if(!strcmp(argv[i], "-go")) GAPOPEN = atof(argv[i+1]); 
    else if(!strcmp(argv[i], "-t")) THREADS = atoi(argv[i+1]); 
    else if(!strcmp(argv[i], "-cores")) CORES = atoi(argv[i+1]); 
//    else if(!strcmp(argv[i], "-noreverse")) REVERSE = 0;
//    else if(!strcmp(argv[i], "-fasta")) { FASTAINPUT = 1; FASTQINPUT = 0; }
//    else if(!strcmp(argv[i], "-nosam")) SAMOUTPUT = 0;
//    else if(!strcmp(argv[i], "-fastq")) { FASTQINPUT = 1; FASTAINPUT = 0; }
    else if(!strcmp(argv[i], "-h")) { printf(help_message); exit(EXIT_SUCCESS); }
    else if(argv[i][0] == '-') { printf("Option %s not defined\n", argv[i]); exit(EXIT_FAILURE); }
  }   
  if(qfilename[0] == '\0' || dbfilenames[0][0] == '\0') {
    printf("Error: please fastq read file (-q) and genome fasta files (-f)\n");
    exit(EXIT_FAILURE);
  }
  if(FRAGMENTSIZE <= 0 || NUMQUERIES <= 0) { 
    printf("Error: please specify positive fragment length (-fl) and number of reads (-nq) \n");
    exit(EXIT_FAILURE);
  }
/*  if(NUMQUERIES == 0) { 
    printf("Error: please specify positive number of reads (-nq) \n");
    exit(EXIT_FAILURE);
  } */

//We need to get number of chars in genome or database to define NUMFRAGMENTS and other variables
  genomelength = get_dblength(genomes, dbfilenames, FRAGMENTSIZE);

//Set QLEN to multiple of 16 and pad later with N's
  assert(FRAGMENTSIZE % 16 == 0);
  numfragments = (int) ( (genomelength + FRAGMENTSIZE - 1) / FRAGMENTSIZE );

  NUMFRAGMENTS = numfragments;

  JOBS = NUMFRAGMENTS;

//Accounting for newline and end of string (null) characters
  int QNAMEBUFFERLEN = QNAME + 2;
  assert(BUFFERLEN >= MAXQLEN);
  char* buffer = (char*) malloc(BUFFERLEN * sizeof(char));

//Allocate space for local variables
  float* results = (float*) malloc(2 * JOBS * sizeof(float));
  char* queryname = (char*) malloc(QNAMEBUFFERLEN * sizeof(char));
  char* query = (char*) malloc(MAXQLEN * sizeof(char));
  float* phred_probs = (float*) malloc(MAXQLEN * sizeof(float));
  char* phred = (char*) malloc(MAXQLEN * sizeof(char));

//We use size_t to avoid integer overflow error
  char* database = (char*) malloc((size_t)JOBS * FRAGMENTSIZE * sizeof(char));
  if(database == NULL) { printf("Error making space for database\n"); exit(EXIT_FAILURE); }

//Make space for matrices for final alignment
//This variable is an upper bound. The actual finalfragment size is smaller and depends upon
//the query length which is not available until we get to the main loop below.
  FINALFRAGMENTSIZE = FRAGMENTSIZE+1;
  char** traceback=(char **) malloc((MAXQLEN+1) * sizeof(char*) );
  float** V=(float **) malloc((MAXQLEN+1) * sizeof(float*) );
  for(i=0;i<(MAXQLEN+1); i++) {
    traceback[i] = (char*) malloc((FINALFRAGMENTSIZE+1) * sizeof(char));
    V[i] = (float*) malloc((FINALFRAGMENTSIZE+1) * sizeof(float));
  }

  float go = GAPOPEN;

//Initialize V and traceback for final alignment
  for(i=0; i <= MAXQLEN; i++){ V[i][0] = 0; traceback[i][0] = 'U'; }
  for(j=0; j <= FINALFRAGMENTSIZE; j++ ){ V[0][j] = 0; traceback[0][j] = 'L'; }

//Make space for final alignment
  firstalign = (char*) malloc((MAXQLEN + FINALFRAGMENTSIZE + 2) * sizeof(char));
  cigar = (char*) malloc((MAXQLEN + FINALFRAGMENTSIZE + 2) * sizeof(char));
  cigarstr = (char*) malloc(MAXQLEN * sizeof(char));
  secondalign = (char*) malloc((MAXQLEN + FINALFRAGMENTSIZE + 2) * sizeof(char));

//Initialize CPU variables
  memset(database, '\0', (size_t)JOBS * FRAGMENTSIZE * sizeof(char));
  memset(results, NEG_INF, 2 * JOBS * sizeof(float));
  memset(query, '\0', MAXQLEN * sizeof(char));

  char contignames[NUMFRAGMENTS * FNAME];
  unsigned int contigposition[NUMFRAGMENTS];

  get_database(database, contignames, contigposition, genomes, dbfilenames, FRAGMENTSIZE, NUMFRAGMENTS, FNAME);

if(OUTPUTTIME){
  gettimeofday(&endtime, NULL); seconds=((double)endtime.tv_sec+(double)endtime.tv_usec/1000000)-((double)starttime.tv_sec+(double)starttime.tv_usec/1000000);}
if(OUTPUTTIME)
   printf("Time to adjust database is %f\n", seconds);

//Print parameters
  if(!SAMOUTPUT){
    printf("Running MaxSSMap_GPU with parameters: %s\n", cmdline);
  }
  else {
    printf("@HD\tVN:1.0\tSO:unsorted\n");
    printf("@PG\tID:maxssmap\tPN:Maxssmap_GPU_int4\tVN:1.0\tCL:%s\n", cmdline);
  }

//Compute sub_matrix and subst_index
  int basecount = strlen(bases);
  for (i = 0; i < basecount; i++)
    for (j = 0; j <= i; j++) {
      if(bases[i]==bases[j])        { sub_matrix[i][j] = 5; }
      else                               { sub_matrix[i][j] = -4; }
      sub_matrix[j][i] = sub_matrix[i][j];
    }

if(DEBUG > 1) {
  for (i = 0; i < basecount; i++){ for (j = 0; j < basecount; j++) printf("%f ", sub_matrix[i][j]); printf("\n"); }
}

  for (i = 0; i < LETTERS; i++) subst_index[i] = -1;
  for (i = 0; i < basecount; i++) subst_index[bases[i] - 'A'] = i;
  for (i = 0; i < PSS_OFFSET; i++) pos_specific_scores[i] = 0;

  char* bestname = (char*) malloc(FNAME * sizeof(char));
  char* bestname2 = (char*) malloc(FNAME * sizeof(char));

  char* bestfragment = (char*) malloc((FINALFRAGMENTSIZE+1) * sizeof(char));
  char* bestfragment2 = (char*) malloc((FINALFRAGMENTSIZE+1) * sizeof(char));
  for(i = 0; i < FINALFRAGMENTSIZE; i++) bestfragment[i] = 'N'; bestfragment[FINALFRAGMENTSIZE]='\0';
  for(i = 0; i < FINALFRAGMENTSIZE; i++) bestfragment2[i] = 'N'; bestfragment2[FINALFRAGMENTSIZE]='\0';

if(!SAMOUTPUT)
  printf("Total reads = %d\n", NUMQUERIES);
if(DEBUG > 0) fflush(stdout);

  assert(FRAGMENTSIZE % 16 == 0);

  char *line = NULL; size_t len = 0; ssize_t read;
  qfile = fopen(qfilename, "r");

//  for(l = 0; l < totalqueries; l++){
  for(l = 0; l < NUMQUERIES; l++){

if(OUTPUTTIME) gettimeofday(&starttime, NULL);
//Reset variables
    bestscore=0; bestscore2=0; 
    strcpy(bestname, blank); strcpy(bestname2, blank); reverse = 0; reverse2 = 0;
//Obtain next query and basequals
//-1 in the assert because the null terminator is not read
    read = getline(&line, &len, qfile); 
//If end of file then exit loop
    if(read == -1) break;
    assert(read < QNAMEBUFFERLEN-1); strcpy(queryname, line); queryname[read-1] = '\0'; 
					//-1 because of the newline
    read = getline(&line, &len, qfile); assert(read -1 < MAXQLEN); strcpy(query, line); query[read-1] = '\0';
    read = getline(&line, &len, qfile);
    read = getline(&line, &len, qfile); assert(read -1 < MAXQLEN); strcpy(phred, line); phred[read-1] = '\0'; 
//Remove newline character
//    query[strlen(query)-2] = '\0'; phred[strlen(phred)-2] = '\0'; queryname[strlen(queryname)-2] = '\0';
    QLEN=strlen(query);

//Fix query and phred and its length by replacing newline with null
    if(QLEN % 16) { QUERY_PADLENGTH = ((int)(QLEN/16) + 1) * 16 - QLEN; QLEN = ((int)(QLEN/16) + 1) * 16; }
    else          { QUERY_PADLENGTH = 0; }
//Make sure query length is within bounds after extending it to multiple of 16
//We use < instead of <= because we keep one extra char for null terminator
    assert(QLEN < MAXQLEN);
//Pad query and phred with extra N's to make multiple of 16
    for(i = strlen(query); i<QLEN; i++) query[i] = 'N';
    query[QLEN] = '\0';
    for(i = strlen(phred); i<QLEN; i++) phred[i] = 'N';
    phred[QLEN] = '\0';
    if(FASTQINPUT) {
      for(i=0; i<strlen(phred); i++)
        if(query[i] == 'N') phred_probs[i] = .25;
        else phred_probs[i] = 1-pow(10, (-1*(float)((int)phred[i]-33)/10));
      phred_probs[i] = '\0';
if(DEBUG > 1) {
      for(i=0; i<strlen(phred); i++) printf("%f ", phred_probs[i]); printf("\n");
}
    }

//if(!SAMOUTPUT) { printf("\nQUERY %s %s (QLEN=%d) ", queryname, query, QLEN); printf("PHRED %s\n", phred);  }

//Compute scorez_matrix and copy to GPU from CPU memory
    compute_scorez_matrix(query, phred, phred_probs);

    dblen = strlen(database);
if(DEBUG > 0)
    printf("padded database length=%u\n", dblen);

if(DEBUG > 0)
    printf("JOBS=%d\n", JOBS);

#pragma omp parallel num_threads(JOBS)
    maxssmap_openmp(JOBS, database, results, FRAGMENTSIZE, QLEN);

//Move to latter half to make space for reverse complement
    for(j = 0; j < JOBS; j++) results[JOBS+j] = results[j];

//And print results
if(DEBUG > 0) {
    for(j = JOBS; j < JOBS+5; j++) printf("%.2f ", results[j]); printf("\n");
    for(j = 2*JOBS-5; j < 2*JOBS; j++) printf("%.2f ", results[j]); printf("\n");
}

//Compute reverse complement of the query
    strcpy(buffer, query);
    for (i = 0; i < basecount; i++) subst_index[bases[i] - 'A'] = i;
    j = 0; for(i = strlen(buffer)-1; i>-1; i--) query[j++] = complement[subst_index[buffer[i] - 'A']];
    query[j]='\0';
    rev(phred);

    if(FASTQINPUT){ 
      for(i=0; i<strlen(query); i++)
        if(query[i] == 'N') phred_probs[i] = .25;
        else phred_probs[i] = 1-pow(10, (-1*(float)((int)phred[i]-33)/10));
      phred_probs[i] = '\0';
if(DEBUG > 1) {
      for(i=0; i<strlen(query); i++) printf("%f ", phred_probs[i]); printf("\n");
}
    }

//if(!SAMOUTPUT) {  printf("REVERSE-COMPLEMENT-QUERY %s (QLEN=%d) ", query, QLEN); printf("PHRED %s\n", phred);  }
    
    compute_scorez_matrix(query, phred, phred_probs);

if(DEBUG > 0) fflush(stdout);

#pragma omp parallel num_threads(JOBS)
    maxssmap_openmp(JOBS, database, results, FRAGMENTSIZE, QLEN);

if(DEBUG > 0){
    for(j = 0; j < 5; j++) printf("%.2f ", results[j]); printf("\n");
    for(j = JOBS-5; j < JOBS; j++) printf("%.2f ", results[j]); printf("\n");
}
    
//Find best score and fragment
    for(j = 0; j < 2*JOBS; j++) 
      if(results[j] > bestscore) { bestscore = results[j]; bestfragmentnumber = j; }
    for(j = 0; j < 2*JOBS; j++) 
      if(j != bestfragmentnumber && results[j] > bestscore2) { bestscore2 = results[j]; bestfragmentnumber2 = j; }
//    if(bestfragmentnumber >= JOBS) { reverse=1; bestfragmentnumber -= JOBS; }
//    if(bestfragmentnumber2 >= JOBS) { reverse2=1; bestfragmentnumber2 -= JOBS; }
    if(bestfragmentnumber >= JOBS) { reverse=0; bestfragmentnumber -= JOBS; } else { reverse = 1; }
    if(bestfragmentnumber2 >= JOBS) { reverse2=0; bestfragmentnumber2 -= JOBS; } else { reverse2 = 1; }

//For comparison against CUDASW++
if(JUSTFRAGMENT) { 
    printf("queryname=%s bestscore=%f bestscore2=%f bestfragmentnumber=%d bestfragmentnumber2=%d\n", queryname, bestscore, bestscore2, bestfragmentnumber, bestfragmentnumber2);
    continue;
}   

    if(!reverse) {
//Unreverse complement the query
      strcpy(buffer, query);
      for (i = 0; i < basecount; i++) subst_index[bases[i] - 'A'] = i;
      j = 0; for(i = strlen(buffer)-1; i>-1; i--) query[j++] = complement[subst_index[buffer[i] - 'A']];
      query[j]='\0';
      if(FASTQINPUT) {
//Unreverse the phred probabilities and recompute scorez_matrix
        rev(phred);
        for(i=0; i<strlen(query); i++)
          if(query[i] == 'N') phred_probs[i] = .25;
          else phred_probs[i] = 1-pow(10, (-1*(float)((int)phred[i]-33)/10));
        phred_probs[i] = '\0';
      }
      compute_scorez_matrix(query, phred, phred_probs);
    }

    strcpy(bestname, &contignames[bestfragmentnumber*FNAME]);
    strcpy(bestname2, &contignames[bestfragmentnumber2*FNAME]);

if(DEBUG > 0) {
    printf("bestscore=%f bestscore2=%f bestfragmentnumber=%d bestfragmentnumber2=%d bestname=%s bestname2=%s\n",
  	    bestscore, bestscore2, bestfragmentnumber, bestfragmentnumber2, bestname, bestname2);
}

    if(DEBUG > 0) fflush(stdout);
//If found a new best fragment
    strcpy(bestname, &contignames[bestfragmentnumber*FNAME]);

    for(ui=0; ui<FRAGMENTSIZE-1; ui++) bestfragment[ui] = database[bestfragmentnumber*FRAGMENTSIZE + ui];
    bestfragment[ui]='\0';

    if(DEBUG > 0) printf("leftfrags=%d rightfrags=%d\n", leftfrags, rightfrags);

    if(DEBUG > 0) fflush(stdout);


//Perform final alignment
    fragmentsize = strlen(bestfragment);
//    assert(fragmentsize == FINALFRAGMENTSIZE);
    assert(fragmentsize < FINALFRAGMENTSIZE);
if(DEBUG > 0)
    printf("Best fragment size = %d\n", fragmentsize);
if(DEBUG > 0)
    printf("Best fragment = %s\n", bestfragment);

if(OUTPUTTIME)
    gettimeofday(&start2time, NULL);

//Doing final NW alignment
    char ch;
    for(i=1; i<=QLEN; i++)
      for(j=1; j<=fragmentsize; j++) {
        ch = bestfragment[j-1]; 
        p1 = V[i-1][j-1] + pos_specific_scores[(i-1)*BASES + subst_index[ch - 'A'] + PSS_OFFSET];  
	if(i == QLEN || j == fragmentsize) { p2 = V[i-1][j]; p3 =  V[i][j-1]; }
        else 				   { p2 = V[i-1][j] + go; p3 = V[i][j-1] + go; }
//        p2 = V[i-1][j] + go; p3 = V[i][j-1] + go;
        if(p1>=p2 && p1>=p3){ V[i][j]=p1; traceback[i][j]='D'; } 
        else if(p2>=p1 && p2>=p3) { V[i][j]=p2; traceback[i][j]='U'; }
        else {  V[i][j]=p3; traceback[i][j]='L';  }
      }

if(OUTPUTTIME){
    gettimeofday(&end2time, NULL); 
    seconds=((double)end2time.tv_sec+(double)end2time.tv_usec/1000000)-((double)start2time.tv_sec+(double)start2time.tv_usec/1000000);
}
if(OUTPUTTIME)
    printf("Time to final DP is %f\n", seconds);

if(OUTPUTTIME)
    gettimeofday(&start2time, NULL);

    firstcounter = QLEN; secondcounter = fragmentsize;
    lencounter = firstcounter + secondcounter;
    firstalign[lencounter-1]='\0'; secondalign[lencounter-1]='\0';
    cigar[lencounter-1]='\0';
    startpos=-1; endpos = -1;
    while( (firstcounter != 0 || secondcounter != 0) ) {
     if( traceback[firstcounter][secondcounter] == 'D') {
        firstalign[lencounter-2] = query[firstcounter-1]; secondalign[lencounter-2] = bestfragment[secondcounter-1];
	cigar[lencounter-2] = 'M';
        if(endpos == -1) { endpos = secondcounter; endalign=lencounter-1; }
        startalign=lencounter-2; startpos = secondcounter; 
        firstcounter--; secondcounter--; 
      }
      else if(traceback[firstcounter][secondcounter] == 'U' ) {
        firstalign[lencounter-2] = query[firstcounter-1]; secondalign[lencounter-2] = '-'; cigar[lencounter-2] = 'I'; firstcounter--;
        if(endpos == -1) { endpos = secondcounter; endalign=lencounter-1; }
        startalign=lencounter-2; startpos = secondcounter; 
      } 
      else {
        secondalign[lencounter-2] = bestfragment[secondcounter-1]; firstalign[lencounter-2] = '-'; cigar[lencounter-2] = 'D'; secondcounter--;
      } 
//      if(endpos == -1) { endpos = secondcounter; endalign=lencounter-1; }
//      startpos = secondcounter; startalign=lencounter-2;  
      lencounter--; 
    } 

if(OUTPUTTIME){
    gettimeofday(&end2time, NULL); 
    seconds=((double)end2time.tv_sec+(double)end2time.tv_usec/1000000)-((double)start2time.tv_sec+(double)start2time.tv_usec/1000000);
}
if(OUTPUTTIME)
    printf("Time to final traceback is %f\n", seconds);

if(OUTPUTTIME)
    gettimeofday(&start2time, NULL);
    cigar[endalign]='\0';
    secondalign[endalign]='\0';
    for(i=0;i<strlen(cigarstr);i++)cigarstr[i]='\0';
    count=0; 
    for(i=startalign; i<endalign; i++) { 
      count++; 
      if(i>startalign && cigar[i] != cigar[i-1]) { 
	sprintf(buffer, "%d", count); 
	strcat(cigarstr, buffer);
	strncat(cigarstr, &cigar[i-1], 1); count=0;
      }
    } 
    sprintf(buffer, "%d", count); 
    strcat(cigarstr, buffer);
    strncat(cigarstr, &cigar[i-1], 1); count=0;
    ch='\0'; strncat(cigarstr, &ch, 1); 
if(OUTPUTTIME){
    gettimeofday(&end2time, NULL); 
    seconds=((double)end2time.tv_sec+(double)end2time.tv_usec/1000000)-((double)start2time.tv_sec+(double)start2time.tv_usec/1000000);
}
if(OUTPUTTIME)
    printf("Time to make cigar string is %f\n", seconds);

if(!SAMOUTPUT) {
//if(1) {
    printf("Reverse=%d Bestname=%s Bestscore=%f Bestfragmentnumber=%d Reverse2=%d Bestname2=%s Bestscore2=%f Bestfragmentnumber2=%d\n", reverse, bestname, bestscore, bestfragmentnumber, reverse2, bestname2, bestscore2, bestfragmentnumber2);
    printf("startpos=%d endpos=%d startfragnumber=%d ", startpos, endpos, bestfragmentnumber);
    for(i=startalign; i<endalign; i++) { printf("%c", firstalign[i]); } printf(" "); for(i=startalign; i<endalign; i++) { printf("%c", secondalign[i]); } printf(" ");
    matches=0; mismatches=0; gaps=0; 
    for(i=startalign; i<endalign; i++) { if(firstalign[i] == 'N' || secondalign[i] == 'N') ; else if(firstalign[i] == secondalign[i]) matches++; else if(firstalign[i] == '-' || secondalign[i] == '-') gaps++; else mismatches++; } 
    for(i=startalign; i<endalign; i++) printf("%c", cigar[i]); printf(" ");
    printf("matches=%f ", matches); printf("mismatches=%f ", mismatches); printf("gaps=%f\n", gaps);
}
else {
    if (reverse == 0) { samout = 0; QUERY_PADLENGTH = 0; } else samout = 16;
    printf("%s\t%d\t%s\t%u\t%f\t%s\t*\t0\t0\t%s\t%s\n", queryname, samout, bestname, contigposition[bestfragmentnumber]-leftfrags*FRAGMENTSIZE+startpos+QUERY_PADLENGTH, -10*log(bestscore2/bestscore)/2.302585092994046, cigarstr, query, phred);
}

    if(DEBUG > 0) fflush(stdout);

if(OUTPUTTIME){
    gettimeofday(&endtime, NULL); 
    seconds=((double)endtime.tv_sec+(double)endtime.tv_usec/1000000)-((double)starttime.tv_sec+(double)starttime.tv_usec/1000000);
}
if(OUTPUTTIME)
    printf("Time to align a query is %f\n", seconds);
  }
 
  free(buffer);
  free(results);
  free(queryname);
  free(phred_probs);
  free(query);
  free(phred);

//  free(querynames);
//  free(queries);
//  free(basequals);

  for(i=0;i<(MAXQLEN+1); i++) {
    free(traceback[i]);
    free(V[i]);
  }
  free(traceback); free(V);
  
  free(firstalign);
  free(cigar);
  free(cigarstr);
  free(secondalign);

  free(bestname);
  free(bestname2); 
  free(bestfragment); 
  free(bestfragment2); 

  return 0;
}

char* rev(char* str) {
  int end= strlen(str)-1;
  int start = 0;

  while( start<end )
  {
    str[start] ^= str[end];
    str[end] ^=   str[start];
    str[start]^= str[end];

    ++start;
    --end;
  }

  return str;
}
